from tokenizer import tokenize_iter
import re
from itertools import ifilter, izip, chain, imap
from tokenizer import make_scanner
from lr import *

lr_grammar_scanner = make_scanner(
    sep='=', alt='[|]', word=r"\b\w+\b", whitespace=r'[ \t\r\n]+',
    minus=r'[-]', discard_names=('whitespace',))


def rules(start, grammar, kw):
    words = [start]
    edit_rule = '@'
    edit_rule_commit = True
    next_edit_rule_commit = True
    kw.add(edit_rule)
    for tokname, tokvalue, tokpos in lr_grammar_scanner(grammar):
        if tokname == 'minus':
            next_edit_rule_commit = False
        if tokname == 'word':
            words.append(tokvalue)
            kw.add(tokvalue)
        elif tokname == 'alt':
            yield (edit_rule, tuple(words), edit_rule_commit)
            words = []
        elif tokname == 'sep':
            tmp = words.pop()
            yield (edit_rule, tuple(words), edit_rule_commit)
            edit_rule_commit = next_edit_rule_commit
            next_edit_rule_commit = True
            edit_rule = tmp
            words = []
    yield (edit_rule, tuple(words), edit_rule_commit)


class RuleSet(dict):

    def __init__(self, rules):
        dict.__init__(self)
        self.init(rules)

    def fill(self, rules):
        self.names_count = 0
        self.rules_count = 0
        self.clear()
        epsilons = set()
        for rulename, elems, commit in rules:
            if len(elems) > 0:
                self.add(rulename, elems, commit)
            else:
                epsilons.add(rulename)
        #print 'found epsilon rules', epsilons
        return epsilons

    def add_epsilon_free(self, eps, epsilons):
        #print "Adding", eps, "-free variants"
        i = 0
        while i < self.rules_count:
            if self[i] is None:
                i += 1
                continue
            rulename, elems, commit = self[i]
            if eps in elems:
                #print "... to", rulename, elems
                E = set([elems])
                old = 0
                while len(E) != old:
                    old = len(E)
                    E = E.union(elems[:i] + elems[i + 1:]
                                for elems in E
                                for i in xrange(len(elems))
                                 if elems[i] == eps)
                #print "Created variants", E
                for elems in E:
                    if len(elems) == 0:
                        #print "got new epsilon rule", rulename
                        epsilons.add(rulename)
                    else:
                        self.add(rulename, elems, commit)
                    #
                #
            i += 1
        #

    def remove_epsilon(self, eps, epsilons):
        must_cleanup = False
        i = 0
        while i < self.rules_count:
            if self[i] is None:
                i += 1
                continue
            rulename, elems, commit = self[i]
            if eps in elems:
                elems = tuple(e for e in elems if e != eps)
                if len(elems) == 0:
                    # yet another epsilon :/
                    self[i] = None
                    self[rulename].remove(i)
                    if not self[rulename]:
                        del self[rulename]
                    must_cleanup = True
                    epsilons.add(rulename)
                    #print "epsilon removal created new epsilon rule", rulename
                else:
                    self[i] = (rulename, elems, commit)
                #
            i += 1
        return must_cleanup

    def init(self, rules):
        epsilons = self.fill(rules)
        must_cleanup = False
        while epsilons:
            eps = epsilons.pop()
            if eps in self:
                # Rule produces something and has an epsilon alternative
                self.add_epsilon_free(eps, epsilons)
            else:
                must_cleanup |= self.remove_epsilon(eps, epsilons)
        if must_cleanup:
            rules = sorted(self[i] for i in xrange(self.rules_count)
                                    if self[i] is not None)
            epsilons = self.fill(rules)
            if epsilons:
                #print "D'oh ! I left epsilon rules in there !", epsilons
                raise Exception("There is a bug ! There is a bug ! " +
                                "Failed to refactor this grammar into " +
                                "an epsilon-free one !")

    def add(self, rulename, elems, commit):
        if rulename not in self:
            self.names_count += 1
            self[rulename] = set()
        rule = (rulename, elems, commit)
        if rule not in (self[i] for i in self[rulename]):
            self[rulename].add(self.rules_count)
            self[self.rules_count] = rule
            self.rules_count += 1


class parser(object):

    def __init__(self, start_sym, grammar, scanner_kw=[]):
        self.kw_set = set(scanner_kw)
        self.kw_set.add('$')
        self.R = RuleSet(rules(start_sym, grammar, self.kw_set))
        self.I = set((r, i) for r in xrange(self.R.rules_count)
                            for i in xrange(len(self.R[r][1]) + 1))
        self.precompute_next_items()
        self.compute_lr0()
        self.LR0 = list(sorted(self.LR0))
        self.LR0_idx = {}
        for i, s in enumerate(self.LR0):
            self.LR0_idx[s] = i
        self.initial_state = self.index(self.initial_items)
        self.compute_ACTION()

    def __str__(self):
        return '\n'.join(self.R[r][0] + ' = ' + ' '.join(self.R[r][1])
                         for r in xrange(self.R.rules_count))

    def conflicts(self):
        "Returns the list of conflicts in the ACTION table."
        return filter(lambda (i, t): len(self.ACTION[i][t]) > 1,
                      ((i, t) for i, row in enumerate(self.ACTION)
                         for t in row.iterkeys()))

    def count_conflicts(self):
        "Returns the count of conflicts in the ACTION table."
        return reduce(lambda a, b: a + (len(b) > 1 and 1 or 0),
                      (a for row in self.ACTION for a in row.itervalues()),
                      0)

    def resolve_SR_conflicts(self, favor='S'):
        for s, k in self.conflicts():
            actions = self.ACTION[s][k]
            atypes = [a[0] for a in actions]
            if 'S' in atypes and 'R' in atypes:
                self.ACTION[s][k] = [a for a in actions if a[0] == favor]

    def compute_lr0(self):
        "Compute the LR(0) sets."
        self.LR0 = set()
        x = closure([(0, 0)], self.R)
        self.initial_items = x
        stack = [tuple(sorted(x))]
        while stack:
            x = stack.pop()
            self.LR0.add(x)
            F = follow(x, self.R)
            for t, s in F.iteritems():
                s = tuple(sorted(s))
                if s not in self.LR0:
                    stack.append(s)

    def itemstr(self, item):
        "Stringify an item for pretty-print."
        return itemstr(item, self.R)

    def itemsetstr(self, item, label=''):
        "Stringify an item set for pretty-print."
        return itemsetstr(item, self.R, label)

    def closure(self, s):
        "Compute the closure of an item set."
        return tuple(sorted(closure(s, self.R)))

    def kernel(self, s):
        "Compute the kernel of an item set."
        return kernel(s)

    def index(self, s):
        """Returns the index of (the closure of) item set s in the LR(0) sets
        list."""
        return self.LR0_idx[self.closure(s)]

    def compute_GOTO(self):
        "Compute the GOTO table."
        self.GOTO = []
        for s in self.LR0:
            f = {}
            for tok, dest in follow(s, self.R).iteritems():
                f[tok] = self.LR0_idx[self.closure(dest)]
            self.GOTO.append(f)

    def init_row(self, init=None):
        "Initialize a row of the ACTION table."
        if init is None:
            init = []
        ret = {}
        for kw in self.kw_set:
            ret[kw] = [] + init
        return ret

    def precompute_next_items(self):
        self.next_list = dict((k, set()) for k in self.R if type(k) is str)
        for item in self.I:
            r, i = item
            n, e, c = self.R[r]
            if i > 0 and e[i - 1] in self.next_list:
                self.next_list[e[i - 1]].add(item)

    def next_items(self, item, visited=None):
        "Compute the yet unvisited items following the given item."
        items = set()
        if visited is None:
            visited = set()
        name = self.R[item[0]][0]
        for it in self.next_list[name]:
            if it not in visited:
                r, i = it
                e = self.R[r][1]
                visited.add(it)
                if len(e) == i:
                    items.update(self.next_items(it, visited))
                else:
                    items.add(it)
        return items

    def following_tokens(self, item):
        "Returns all tokens following the current item."
        items = self.next_items(item)
        ret = first(closure(items, self.R), self.R)
        ret.add('$')
        return ret

    def compute_ACTION(self):
        "Compute the ACTION/GOTO table."
        self.compute_GOTO()
        self.ACTION = []
        for s, g in izip(self.LR0, self.GOTO):
            action = self.init_row()
            # reductions
            for r, i in ifilter(lambda (r, i): i == len(self.R[r][1]), s):
                if not r:
                    action['$'].append(('A',))
                else:
                    for kw in self.following_tokens((r, i)):
                    #for kw in self.kw_set:
                        action[kw].append(('R', r))
            # transitions
            for tok, dest in g.iteritems():
                action[tok].append(('S', dest))
            # commit
            self.ACTION.append(action)

    def action_to_str(self):
        "Stringify the ACTION/GOTO table for pretty-print."

        def ac_str(c):
            return ''.join(imap(str, c))

        def cell(i, kw):
            if i >= 0:
                return ','.join(map(ac_str, self.ACTION[i][kw]))
            if i < 0:
                return kw != '@' and str(kw) or ''

        def col_width(kw):
            return reduce(max, chain([type(kw) is str and len(kw) or kw],
                                     (len(cell(i, kw))
                                      for i in xrange(len(self.ACTION)))))

        col_labels = sorted(self.kw_set,
                            key=lambda x: x in self.R and '|' + x
                                          or x == '$' and '|$'
                                          or x)
        col_widths = [kw != '@' and col_width(kw) or 0 for kw in col_labels]

        def row(i):
            return ' | '.join(cell(i, kw).center(cw)
                              for kw, cw in izip(col_labels, col_widths))

        header = '    | ' + row(-1) + '\n'
        return header + '\n'.join(('%3i | ' % i) + row(i)
                                  for i in xrange(len(self.ACTION)))

    def dump_sets(self):
        "Pretty-print all LR(0) item sets."
        for i, lrset in enumerate(self.LR0):
            print self.itemsetstr(lrset, i)
            print

    def itemset(self, i):
        return self.itemsetstr(self.LR0[i], i)

    @property
    def unused_rules(self):
        check = lambda i: reduce(lambda a, b: a and i not in b, self.LR0, True)
        unused_rule_indices = set(x[0] for x in filter(check, self.I))
        return set(self.R[x][0] for x in unused_rule_indices)



__all__ = ['stack_item', 'stack']

from parser import parser
from itertools import chain, ifilter


def printlist(l, pfx):
    print pfx + '\n  ' + '\n  '.join(map(str, l))


class stack_item(object):

    def __init__(self, prev, x):
        self.prev = prev
        self.data = x

    def __str__(self):
        return "stack_item<%s>" % str(self.data)

    __repr__ = __str__

    #def __hash__(self):
    #    return hash(self.data)

    #def __eq__(self, x):
    #    return self.data.__eq__(x.data)


class stack(object):

    def __init__(self, A):
        self.active = []
        self.A = A
        self.count_active = 0
        self.previously_active = []

    def enumerate_active(self):
        i = 0
        while i < len(self.active):
            yield i, self.active[i]
            i += 1

    def shift(self, source, token, state):
        #printlist(self.rec_all_pathes(source), "before shift: ")
        sit = stack_item(set([source]), token)
        sis = stack_item(set([sit]), state)
        self.active.append(sis)
        #printlist(self.rec_all_pathes(sis), "after shift: ")
        #print "shift", source, token, state, self.active

    def rec_path(self, node, n):
        #print "rec_path(%s, %s)" % (str(node), str(n))
        if n == 0:
            return set(((node,),))
        if not node.prev:
            return None
        prec_pathes = set(path for prev in node.prev
                               for path in self.rec_path(prev, n - 1))
        ret = set(path + (node.data,) for path in prec_pathes)
        return ret

    def rec_all_pathes(self, node):
        if node is None:
            return [[]]
        ret = [path.append(node.data) or path
               for prev in node.prev
               for path in self.rec_all_pathes(prev)]
        return ret

    def reduce(self, node, ruleidx):
        name, elems, commit = self.A.R[ruleidx]
        pathes = self.rec_path(node, len(elems) * 2)
        #print "found", len(pathes), "pathes", pathes
        #print "================================="
        #print 'R' + '\nR'.join(map(str, pathes))
        #print "---------------------------------"
        #tokenlists = set()
        #for path in pathes:
        #    tokenlists.add(tuple(e for el in path[1::2] for e in el))
        #for tokens in tokenlists:
        for path in pathes:
            tokens = tuple(e for el in path[1::2] for e in el)
            if commit:
                ast = tuple(chain([name], tokens))
                ok = self.A.validate_ast(ast)
                if ok is not None:
                    ok = (ok != tuple()) and (ok,) or tuple()
            else:
                ast = tokens
                ok = ast
            if ok is not None:
                goto = self.A.ACTION[path[0].data][name]
                self.shift(path[0], ok, goto[0][1])

    def merge(self):
        merged_s = {}
        self.previously_active = self.active[:self.count_active]
        #printlist(list(p
        #               for sis in self.active[self.count_active:]
        #               for p in self.rec_all_pathes(sis)),
        #          "before merge: ")
        for node in self.active[self.count_active:]:
            state = node.data
            if state in merged_s:
                merged_s[state].prev.update(node.prev)
            else:
                merged_s[state] = node
        self.active = merged_s.values()
        self.count_active = len(self.active)
        #printlist(list(p
        #               for sis in self.active
        #               for p in self.rec_all_pathes(sis)),
        #          "after merge: ")

    def dump(self):
        print "GSS HAS", self.count_active, "ACTIVE STATES"
        for a in self.active:
            print self.A.itemsetstr(self.A.kernel(self.A.LR0[a.data]), a.data)

    def accepts(self):
        AC = self.A.ACTION
        is_accepting = lambda state: len(AC[state]['$']) > 0 \
                                     and AC[state]['$'][0][0] == 'A'
        # Ast's are always encapsulated inside a 1-uple. We don't want that
        # in the output. Hence the [0].
        ret = [path[-2][0] for node in self.active
                            if is_accepting(node.data)
                           for path in self.rec_all_pathes(node)
                            if len(path) == 4]
                           # length 4 being
                           # (None, initial_state, AST, accepting_state)
                           # we want the AST, hence the [-2].
        #print "GSS ACCEPTING", len(ret), "OUT OF", self.count_active, "STATES"
        return ret



__all__ = ['pp_ast', 'Automaton', 'Scanner', 'make_scanner']

from automaton import Automaton
from tokenizer import Scanner, make_scanner


def pp_ast(ast, indent='', is_last=True):
    """Pretty-print an AST node created by the parser."""
    if is_last:
        prefix = '`- '
        follow = '   '
    else:
        prefix = '+- '
        follow = '|  '
    if type(ast) is not tuple:
        print ''.join((indent, prefix, str(ast)))
    elif len(ast) and type(ast[1]) is not str:  # be it a tuple or anything
        print ''.join((indent, prefix, ast[0]))
        subi = indent + follow
        for x in ast[1:-1]:
            pp_ast(x, subi, False)
        pp_ast(ast[-1], subi, True)
    else:
        print ''.join((indent, prefix, str(ast)))



__all__ = ["TokenizerException", "Scanner", "make_scanner", "token_line_col"]

import re
from itertools import chain

# Code adapted from an answer at stackoverflow.com
# (url split in 2 to please pep8)
#http://stackoverflow.com/questions/2358890
#/python-lexical-analysis-and-tokenization


def token_line_col(text, tok):
    """Converts the token offset into (line, column) position.
    First character is at position (1, 1)."""
    line = text.count('\n', 0, tok[2]) + 1
    offset = text.rfind('\n', 0, tok[2])
    if offset == -1:
        column = tok[2] + 1
    else:
        column = tok[2] - offset
    return line, column


class TokenizerException(Exception):
    pass


def tokenize_iter(text, token_re, discard_names={}, discard_values={}):
    pos = 0
    while True:
        m = token_re.match(text, pos)
        if not m:
            break
        pos = m.end()
        tokname = m.lastgroup
        try:
            tokvalue = m.group(tokname)
        except IndexError, ie:
            print "No such group", tokname
        tokpos = m.start()
        if tokname not in discard_names and tokvalue not in discard_values:
            yield tokname, tokvalue, tokpos
    if pos != len(text):
        msg = 'tokenizer stopped at pos %r of %r in "%s" at "%s"' % (
                pos, len(text), text, text[pos:pos + 3])
        raise TokenizerException(msg)
# End of adapted code


check_groups = re.compile('[(][?]P=(\w+)[)]')


class Scanner(object):

    def __init__(self, **tokens):
        self.re = re.compile('')
        self.tokens = {}
        self.discard_names = set()
        self.discard_values = set()
        self.add(**tokens)

    def add(self, **tokens):
        """Each named keyword is a token type and its value is the
        corresponding regular expression. Returns a function that iterates
        tokens in the form (type, value) over a string.

        Special keywords are discard_names and discard_values, which specify
        lists (actually any iterable is accepted) containing tokens names or
        values that must be discarded from the scanner output.
        """
        if 'discard_names' in tokens:
            self.discard_names.update(tokens['discard_names'])
            del tokens['discard_names']
        if 'discard_values' in tokens:
            self.discard_values.update(tokens['discard_values'])
            del tokens['discard_values']
        # Check there is no undefined group in an assertion
        for k, v in tokens.iteritems():
            bad_groups = filter(lambda g: g not in tokens,
                                check_groups.findall(v))
            if bad_groups:
                print "Unknown groups", bad_groups
        pattern_gen = ('(?P<%s>%s)' % (k, v) for k, v in tokens.iteritems())
        if self.re.pattern:
            pattern_gen = chain((self.re.pattern,), pattern_gen)
        self.re = re.compile('|'.join(pattern_gen), re.VERBOSE | re.MULTILINE)
        self.tokens.update(tokens)
        return self

    def __call__(self, text):
        "Iteratively scans through text and yield each token"
        return tokenize_iter(text, self.re,
                             self.discard_names, self.discard_values)


def make_scanner(**tokens):
    return Scanner(**tokens)



from itertools import ifilter, imap


def expand_item(item, R):
    rule = R[item[0]]
    return rule[1], item[1], rule[0]


def expand_itemset(itemset, R):
    return imap(lambda x: expand_item(x, R), itemset)


def expand_itemset2(itemset, R):
    for item in itemset:
        rule = R[item[0]]
        yield item[0], rule[1], item[1], rule[0]
    #return imap(lambda x: x[:1] + expand_item(x, R), itemset)


def itemstr(item, R):
    e, i, n = expand_item(item, R)
    return ("[%s -> %s . %s" %
                (n, ' '.join(e[:i]), ' '.join(e[i:]))).strip() + ']'


def itemsetstr(itemset, R, label=''):
    items = map(lambda x: itemstr(x, R), sorted(itemset))
    width = reduce(lambda a, b: max(a, len(b)), items, 3)
    label = label and '[' + str(label) + ']' or ''
    build = ["+-%s%s-+" % (label, '-' * (width - len(label)))]
    build.extend("| %-*s |" % (width, item) for item in items)
    build.append("+-" + "-" * width + '-+')
    return '\n'.join(build)


def first(itemset, R):
    "set of the tokens at the right of each dot in this item set"
    ret = set()
    for ruleelems, i, rulename in expand_itemset(itemset, R):
        if i == len(ruleelems):
            continue
        e = ruleelems[i]
        if not e in R:
            ret.add(e)
    return ret


def follow(itemset, R):
    "all transitions from an item set in a dictionary [token]->item set"
    ret = dict()
    for ruleidx, ruleelems, i, rulename in expand_itemset2(itemset, R):
        if i == len(ruleelems):
            continue
        e = ruleelems[i]
        if e not in ret:
            ret[e] = set()
        ret[e].update(closure([(ruleidx, i + 1)], R))
    return ret


def closure(itemset, R):
    "the epsilon-closure of this item set"
    C = set(itemset)
    last = -1
    visited = set()
    while len(C) != last:
        last = len(C)
        Ctmp = set()
        for item in C:
            r, i = item
            name, elems, commit = R[r]
            #elems, i, name = expand_item(item, R)
            if i == len(elems):
                continue
            if elems[i] in R and elems[i] not in visited:
                visited.add(elems[i])
                for r in R[elems[i]]:
                    Ctmp.add((r, 0))
                #Ctmp.update((r, 0) for r in R[elems[i]])
        C.update(Ctmp)
    return C


def kernel(itemset):
    "the kernel items in this item set"
    return set(ifilter(lambda (r, i): not r or i, itemset))



#!/usr/bin/env python

from distutils.core import setup

setup(name='jupyLR',
      version='0.3',
      description='GLR parsing for python with on-line ' +
                  'AST validation and basic error reporting.',
      author='Damien "bl0b" Leroux',
      author_email='damien.leroux@gmail.com',
      url='http://bl0b.github.com/jupyLR/',
      packages=['jupyLR'],
     )



__all__ = ['Automaton']

from parser import parser
from itertools import ifilter, chain
from stack import stack
#from lr import kernel
from tokenizer import token_line_col

INITIAL_TOKEN = ('#', '#', 0)


class Automaton(parser):
    """A GLR parser."""

    def __init__(self, start_sym, grammar, scanner):
        """Construct a GLR automaton given an LR grammar and a start symbol.
        The scanner that creates the automaton input is also provided for
        convenience of use."""
        parser.__init__(self, start_sym, grammar, scanner.tokens.keys())
        self.scanner = scanner
        self.debug = False

    def validate_ast(self, ast):
        """Overload this method in subclasses to discriminate semantically
        invalid ASTs on the fly.
        Return None for an invalid AST, or whatever you want to end up in the
        resulting AST.
        """
        return ast

    def error_detected(self, cur_tok, last_states):
        """Overload this method in subclasses to implement error recovery
        or notification.
        """
        line, column = token_line_col(self.text, cur_tok)
        print "Error detected at line %i, column %i:" % (line, column)
        lines = self.text.splitlines()
        print lines and lines[line - 1] or self.text
        print '%s^' % (' ' * (column - 1))
        #for st in last_states:
        #    print self.itemsetstr(kernel(self.LR0[st.data]))
        #    print "Expected", ', '.join(kw
        #                                for kw in self.kw_set
        #                                if len(self.ACTION[st.data][kw]) > 0)
        A = self.ACTION
        toks = set(kw for st in last_states for kw in self.kw_set
                       if len(A[st.data][kw]) > 0
                          and kw not in self.R and kw != '$')
        if not toks:
            print "Expected end of text"
        else:
            print "Expected token%s" % (len(toks) > 1 and 's' or ''),
            print ', '.join(toks)
        if self.debug:
            rules = set(kw for st in last_states for kw in self.kw_set
                            if len(A[st.data][kw]) > 0 and kw in self.R)
            print "Expected rules", ', '.join(rules)
        return False

    def __recognize(self, token_stream):
        """Runs the automaton over an input stream of tokens. For use with the
        output of a Scanner instance. The last token MUST be
        ('$', '$', length_of_text)."""
        S = stack(self)
        #toki = iter(token_stream)
        S.shift(None, None, 0)
        S.count_active = 1
        prev_tok = INITIAL_TOKEN
        for cur_tok in token_stream:
            if len(S.active) == 0:
                if not self.error_detected(prev_tok, S.previously_active):
                    break
                else:
                    continue
            prev_tok = cur_tok
            #print "pre reduce", S.active
            # Reduce phase
            for i, node in S.enumerate_active():  # S.active may grow
                state = node.data
                for r, rule in ifilter(lambda x: x[0] == 'R',
                                       self.ACTION[state][cur_tok[0]]):
                    #print "R", i, len(S.active), node
                    S.reduce(node, rule)
                i += 1
            # Check if there are accepting states, and return their outputs
            if cur_tok[0] == '$':
                acc = S.accepts()
                if acc:
                    return acc
                else:
                    self.error_detected(cur_tok, S.active)
            #print "pre shift", S.active
            # Shift phase
            S.count_active = len(S.active)
            for node in (S.active[i] for i in xrange(len(S.active))):
                state = node.data
                for r, state in ifilter(lambda x: x[0] == 'S',
                                       self.ACTION[state][cur_tok[0]]):
                    S.shift(node, (cur_tok,), state)
            #print "pre merge", S.active
            # Merge states
            S.merge()
            # A little bit of feedback
            if self.debug:
                print "On token", cur_tok
                S.dump()
        return None

    def __call__(self, text):
        """Parse this text and return a list of valid ASTs, if any. On error,
        returns None."""
        self.text = text  # for the sake of the error detection/recovery
        return self.__recognize(chain(self.scanner(text),
                                      [('$', '$', len(self.text))]))
